pragma solidity >=0.4.21 <0.6.0;

import "openzeppelin-solidity/contracts/ownership/Ownable.sol";
import "openzeppelin-solidity/contracts/token/ERC20/IERC20.sol";

/**
 * A base contract for managing deposits into the Vidy ecosystem.  The basic
 * process: transfer VidyCoins into the appropriate deposit account, then inform
 * the API layer that the tokens are available for retrieval.  The API then
 * performs the withdrawal from the deposit account along with updating its
 * own records.
 *
 * Advantages of this approach:
 * 1. Avoid potential abuse of `transferFrom`, the standard way to use ERC20
 *    tokens to pay for contract actions.  In particular, safe use of
 *    `transferFrom` requires careful management of sender addresses -- if there
 *    are multiple reasons why an account might allow withdrawal by our backend,
 *    we have to make sure the correct action results from that withdrawal.
 *    This problem doesn't exist for ERC827 tokens which use `transferAndCall`
 *    instead, but VidyCoin isn't one of those.
 * 2. Clear record of exactly how many new VidyCoins have been deposited: the
 *    intermediate deposit into this contract, from which the coins are retrieved
 *    concurrently with them becoming available to the backend, compared to
 *    directly transfering tokens to the account from which they are spent.
 */
contract BaseDepositAccount is Ownable {

  mapping(uint256 => bool) public nonceUsed;
  mapping(uint256 => Transfer) public transfers;
  uint256 public transferCount;

  mapping(address => uint256) public totalTransferred;

  struct Transfer {
    address token;
    address to;
    uint256 value;
    uint256 nonce;
  }

  event TransferredEther(address indexed to, uint256 value, uint256 indexed nonce);
  event TransferredTokens(address indexed token, address indexed to, uint256 value, uint256 indexed nonce);
  event Transferred(address indexed token, address indexed to, uint256 value, uint256 indexed nonce);

  constructor(

  ) public { }

  function() external payable { }

  function transferEtherToOwner(uint256 _value, uint256 _nonce) public onlyOwner returns (bool) {
    address payable to = address(uint160(owner()));
    return _transfer(address(0), to, _value, _nonce);
  }

  function transferEther(address payable _to, uint256 _value, uint256 _nonce) public onlyOwner returns (bool) {
    return _transfer(address(0), _to, _value, _nonce);
  }

  function transferTokensToOwner(address _token, uint256 _value, uint256 _nonce) public onlyOwner returns (bool) {
    address payable to = address(uint160(owner()));
    return _transfer(_token, to, _value, _nonce);
  }

  function transferTokens(address _token, address payable _to, uint256 _value, uint256 _nonce) public onlyOwner returns (bool) {
    return _transfer(_token, _to, _value, _nonce);
  }

  function tokenBalance(address _token) public view returns (uint256) {
    IERC20 token = IERC20(_token);
    return token.balanceOf(address(this));
  }

  function _transfer(address _token, address payable _to, uint256 _value, uint256 _nonce) internal returns (bool) {
    require(!nonceUsed[_nonce]);

    nonceUsed[_nonce] = true;
    transfers[_nonce] = Transfer({
        to: _to,
        token: _token,
        value: _value,
        nonce: _nonce
    });
    transferCount = transferCount + 1;
    totalTransferred[_token] = totalTransferred[_token] + _value;

    if (_token == address(0)) {
      // transfer ether and emit event
      _to.transfer(_value);
      emit TransferredEther(_to, _value, _nonce);
    } else {
      // get a token interface, transfer tokens, and emit the event
      IERC20 token = IERC20(_token);
      token.transfer(_to, _value);
      emit TransferredTokens(_token, _to, _value, _nonce);
    }

    emit Transferred(_token, _to, _value, _nonce);
    return true;
  }

}
