const BaseDepositAccount = artifacts.require("./BaseDepositAccount.sol");
const TestToken = artifacts.require("./TestToken.sol");

const BigNumber = require('bignumber.js');

const { assertRevert, validateErc20, assertNumbersEqual, assertNumberIsSum, assertNumberIsDifference, assertNumberIsGte, padAddress } = require('@vidy-dev/ethereum-utils')

const tokenSupply = 10000;
const tokenDeposit = 4000;
const tokenAmount = 1500;
const nonce = 1234;
const etherDeposit = BigNumber('4e17').toFixed();
const etherAmount = BigNumber('1.5e17').toFixed();
const etherRemainder = BigNumber('2.5e17').toFixed();
let depositAccount, token;
let depositOwner, tokenHolder;

contract('BaseDepositAccount', (accounts) => {
  beforeEach(async() => {
    depositOwner = accounts[1];
    tokenHolder = accounts[2];

    token = await TestToken.new(tokenSupply);
    depositAccount = await BaseDepositAccount.new({from: depositOwner});

    await token.transfer(tokenHolder, tokenSupply - tokenDeposit);
    await token.transfer(depositAccount.address, tokenDeposit);
    await sendEther(accounts[3], depositAccount.address, etherDeposit);
  });

  it('construction: should remember owner', async() => {
    const obj0 = await BaseDepositAccount.new();
    const obj1 = await BaseDepositAccount.new({from: accounts[1]});
    assert.strictEqual(await obj0.owner.call(), accounts[0]);
    assert.strictEqual(await obj1.owner.call(), accounts[1]);
  });

  it('transferOwnership: should revert for non-owner', async() => {
    const obj0 = await BaseDepositAccount.new();
    const obj1 = await BaseDepositAccount.new({from: accounts[1]});
    await assertRevert(obj0.transferOwnership.call(accounts[2], {from: accounts[1]}));
    await assertRevert(obj1.transferOwnership.call(accounts[2]));
  });

  it('transferOwnership: should succeed for owner', async() => {
    const obj0 = await BaseDepositAccount.new();
    const obj1 = await BaseDepositAccount.new({from: accounts[1]});
    await obj0.transferOwnership(accounts[2]);
    await obj1.transferOwnership(accounts[2], {from: accounts[1]});
    assert.strictEqual(await obj0.owner.call(), accounts[2]);
    assert.strictEqual(await obj1.owner.call(), accounts[2]);
  });

  it('transferEther: should transfer ether successfully', async () => {
    await testTransferEther(etherAmount);
  });

  it('transferEther: should transfer 0 ether successfully', async () => {
    await testTransferEther(0);
  });

  it('transferEther: should transfer 1 ether successfully', async () => {
    await testTransferEther(1);
  });

  it('transferEther: should transfer full ether deposit successfully', async () => {
    await testTransferEther(etherDeposit);
  });

  it('transferEther: should transfer full ether deposit minus 1 successfully', async () => {
    await testTransferEther(BigNumber(etherDeposit).minus(1).toFixed());
  });

  it('transferEther: should record transfer in internal records', async () => {
    await depositAccount.transferEther(accounts[3], etherAmount, 1234, {from: depositOwner});
    await depositAccount.transferEther(accounts[4], 1, 1235, {from: depositOwner});
    await testTransferEtherRecords(accounts[3], etherAmount, 1234);
    await testTransferEtherRecords(accounts[4], 1, 1235);
    assert.strictEqual((await depositAccount.totalTransferred.call(padAddress(0))).toString(), BigNumber(etherAmount).plus(1).toString());
  });

  it('transferEtherToOwner: should transfer ether successfully', async () => {
    await testTransferEtherToOwner(etherAmount);
  });

  it('transferEtherToOwner: should transfer 0 ether successfully', async () => {
    await testTransferEtherToOwner(0);
  });

  it('transferEtherToOwner: should transfer 1 ether successfully', async () => {
    await testTransferEtherToOwner(1);
  });

  it('transferEtherToOwner: should transfer full ether deposit successfully', async () => {
    await testTransferEtherToOwner(etherDeposit);
  });

  it('transferEtherToOwner: should transfer full ether deposit minus 1 successfully', async () => {
    await testTransferEtherToOwner(BigNumber(etherDeposit).minus(1).toFixed());
  });

  it('transferEtherToOwner: should record transfer in internal records', async () => {
    await depositAccount.transferEtherToOwner(etherAmount, 1234, {from: depositOwner});
    await depositAccount.transferEtherToOwner(1, 1235, {from: depositOwner});
    await testTransferEtherRecords(depositOwner, etherAmount, 1234);
    await testTransferEtherRecords(depositOwner, 1, 1235);
    assert.strictEqual((await depositAccount.totalTransferred.call(padAddress(0))).toString(), BigNumber(etherAmount).plus(1).toString());
  });

  it('transferTokens: should transfer ERC20 tokens successfully', async () => {
    await testTransferTokens(tokenAmount);
  });

  it('transferTokens: should transfer 0 ERC20 token successfully', async () => {
    await testTransferTokens(0);
  });

  it('transferTokens: should transfer 1 ERC20 token successfully', async () => {
    await testTransferTokens(1);
  });

  it('transferTokens: should transfer full ERC20 token deposit successfully', async () => {
    await testTransferTokens(tokenDeposit);
  });

  it('transferTokens: should transfer full ERC20 token deposit minus 1 successfully', async () => {
    await testTransferTokens(tokenDeposit - 1);
  });

  it('transferTokens: should record transfer in internal records', async () => {
    await depositAccount.transferTokens(token.address, accounts[3], tokenAmount, 1234, {from: depositOwner});
    await depositAccount.transferTokens(token.address, accounts[4], 1, 1235, {from: depositOwner});
    await testTransferTokenRecords(token, accounts[3], tokenAmount, 1234);
    await testTransferTokenRecords(token, accounts[4], 1, 1235);
    assert.strictEqual((await depositAccount.totalTransferred.call(token.address)).toString(), `${tokenAmount + 1}`);
  });

  it('transferTokensToOwner: should transfer ERC20 tokens successfully', async () => {
    await testTransferTokensToOwner(tokenAmount);
  });

  it('transferTokensToOwner: should transfer 0 ERC20 token successfully', async () => {
    await testTransferTokensToOwner(0);
  });

  it('transferTokensToOwner: should transfer 1 ERC20 token successfully', async () => {
    await testTransferTokensToOwner(1);
  });

  it('transferTokensToOwner: should transfer full ERC20 token deposit successfully', async () => {
    await testTransferTokensToOwner(tokenDeposit);
  });

  it('transferTokensToOwner: should transfer full ERC20 token deposit minus 1 successfully', async () => {
    await testTransferTokensToOwner(tokenDeposit - 1);
  });

  it('transferTokensToOwner: should record transfer in internal records', async () => {
    await depositAccount.transferTokensToOwner(token.address, tokenAmount, 1234, {from: depositOwner});
    await depositAccount.transferTokensToOwner(token.address, 1, 1235, {from: depositOwner});
    await testTransferTokenRecords(token, depositOwner, tokenAmount, 1234);
    await testTransferTokenRecords(token, depositOwner, 1, 1235);
    assert.strictEqual((await depositAccount.totalTransferred.call(token.address)).toString(), `${tokenAmount + 1}`);
  });

  it('transfer* anything: should revert for non-owner', async () => {
    await assertRevert(depositAccount.transferEther.call(accounts[2], etherAmount, 0, {from: tokenHolder}));
    await assertRevert(depositAccount.transferEtherToOwner.call(etherAmount, 0, {from: accounts[0]}));
    await assertRevert(depositAccount.transferTokens.call(token.address, accounts[2], tokenAmount, 0));
    await assertRevert(depositAccount.transferTokensToOwner.call(token.address, tokenAmount, 0, {from: accounts[5]}));
  });

  it('transfer* anything: should revert for amount greater than deposit', async () => {
    const etherDepositPlus = BigNumber(etherDeposit).plus(1).toFixed();
    const tokenDepositPlus = tokenDeposit + 1;
    await assertRevert(depositAccount.transferEther.call(accounts[2], etherDepositPlus, 0, {from: depositOwner}));
    await assertRevert(depositAccount.transferEtherToOwner.call(etherDepositPlus, 0, {from: depositOwner}));
    await assertRevert(depositAccount.transferTokens.call(token.address, accounts[2], tokenDepositPlus, 0, {from: depositOwner}));
    await assertRevert(depositAccount.transferTokensToOwner.call(token.address, tokenDepositPlus, 0, {from: depositOwner}));
  });

  it('transfer* anything: should revert for nonce already used', async () => {
    await depositAccount.transferEther(accounts[2], etherAmount, 0, {from: depositOwner});
    await depositAccount.transferEtherToOwner(etherAmount, 1, {from: depositOwner});
    await depositAccount.transferTokens(token.address, accounts[2], tokenAmount, 2, {from: depositOwner});
    await depositAccount.transferTokensToOwner(token.address, tokenAmount, nonce, {from: depositOwner});

    await assertRevert(depositAccount.transferEther.call(accounts[2], 1, nonce, {from: depositOwner}));
    await assertRevert(depositAccount.transferEtherToOwner.call(1, 2, {from: depositOwner}));
    await assertRevert(depositAccount.transferTokens.call(token.address, accounts[2], 1, 1, {from: depositOwner}));
    await assertRevert(depositAccount.transferTokensToOwner.call(token.address, 1, 0, {from: depositOwner}));
  });

  it('events: should fire Transferred and TransferredEther events for transferEther', async () => {
    const res = await depositAccount.transferEther(accounts[2], etherAmount, nonce, {from: depositOwner});
    await testTransferEtherLogs(res, accounts[2], etherAmount, nonce);
  });

  it('events: should fire Transferred and TransferredEther events for transferEtherToOwner', async () => {
    const res = await depositAccount.transferEtherToOwner(etherAmount, nonce, {from: depositOwner});
    await testTransferEtherLogs(res, depositOwner, etherAmount, nonce);
  });

  it('events: should fire Transferred and TransferredTokens events for transferTokens', async () => {
    const res = await depositAccount.transferTokens(token.address, accounts[2], tokenAmount, nonce, {from: depositOwner});
    await testTransferTokenLogs(res, token, accounts[2], tokenAmount, nonce);
  });

  it('events: should fire Transferred and TransferredTokens events for transferTokensToOwner', async () => {
    const res = await depositAccount.transferTokensToOwner(token.address, tokenAmount, nonce, {from: depositOwner});
    await testTransferTokenLogs(res, token, depositOwner, tokenAmount, nonce);
  });



  /*
  it('erc20: should pass erc20 standard validation', async() => {
    const balances = {};
    balances[accounts[0]] = balance0;
    balances[accounts[1]] = balance1;
    await validateErc20({
      token: T,
      supply,
      decimals: 18,
      name: "Token",
      symbol: "T",
      accounts,
      balances,
      allowances: {}
    })
  });

  it('construction: should remember owner', async() => {
      const obj0 = await BaseERC20Token.new(supply, 18, "Token", "T");
      const obj1 = await BaseERC20Token.new(supply, 18, "Token", "T", {from: accounts[1]});
      assert.strictEqual(await obj0.owner.call(), accounts[0]);
      assert.strictEqual(await obj1.owner.call(), accounts[1]);
  });

  it('pause: should set "paused"', async() => {
    assert.strictEqual(await T.paused.call(), false);
    await T.pause({from: owner});
    assert.strictEqual(await T.paused.call(), true);
  });

  it('pause: all transfers and approvals should revert while paused', async() => {
    await T.approve(accounts[1], 1000, {from: accounts[0]});
    await T.pause({from: accounts[0]});

    await assertRevert(T.transfer(accounts[1], 1, {from: accounts[0]}));
    await assertRevert(T.transfer(accounts[0], 1, {from: accounts[1]}));
    await assertRevert(T.transfer(accounts[1], 0, {from: accounts[0]}));
    await assertRevert(T.transfer(accounts[0], 0, {from: accounts[1]}));

    await assertRevert(T.transferFrom(accounts[0], accounts[2], 1, {from: accounts[1]}));

    await assertRevert(T.approve(accounts[0], 1000, {from: accounts[1]}));
  })

  it('unpause: should set "paused"', async() => {
    assert.strictEqual(await T.paused.call(), false);
    await T.pause({from: owner});
    await T.unpause({from: owner})
    assert.strictEqual(await T.paused.call(), false);
  });

  it('unpause: should pass erc20 standard validation after unpausing', async() => {
    const balances = {};
    balances[accounts[0]] = balance0;
    balances[accounts[1]] = balance1;
    await T.pause({from: owner});
    await T.unpause({from: owner})
    await validateErc20({
      token: T,
      supply,
      decimals: 18,
      name: "Token",
      symbol: "T",
      accounts,
      balances,
      allowances: {}
    })
  })

  it('burn: should successfully burn tokens, removing them from total supply', async() => {
    await T.burn(10, { from: accounts[0] });
    const balance = await T.balanceOf.call(accounts[0]);
    const totalSupply = await T.totalSupply.call();

    assert.strictEqual(balance.toNumber(), balance0 - 10);
    assert.strictEqual(totalSupply.toNumber(), supply - 10);
  });

  it('burn: revert when paused', async() => {
    await T.pause({from: owner});
    await assertRevert(T.burn(10, { from: accounts[0] }));
    await assertRevert(T.burn(10, { from: accounts[1] }));
  });

  it('burn: should successfully burn tokens after unpausing', async() => {
    await T.pause({from: owner});
    await T.unpause({from: owner});
    await T.burn(10, { from: accounts[0] });
    const balance = await T.balanceOf.call(accounts[0]);
    const totalSupply = await T.totalSupply.call();

    assert.strictEqual(balance.toNumber(), balance0 - 10);
    assert.strictEqual(totalSupply.toNumber(), supply - 10);
  });

  it('burn: should successfully burn all tokens in an account', async() => {
    await T.burn(balance0, { from: accounts[0] });
    const balance = await T.balanceOf.call(accounts[0]);
    const totalSupply = await T.totalSupply.call();

    assert.strictEqual(balance.toNumber(), 0);
    assert.strictEqual(totalSupply.toNumber(), supply - balance0);
  });

  it('burn: should successfully 0 tokens', async() => {
    await T.burn(0, { from: accounts[0] });
    const balance = await T.balanceOf.call(accounts[0]);
    const totalSupply = await T.totalSupply.call();

    assert.strictEqual(balance.toNumber(), balance0);
    assert.strictEqual(totalSupply.toNumber(), supply);
  });

  it('burn: should fail burning more tokens than in an account', async() => {
    await assertRevert(T.burn.call(balance0 + 1, { from: accounts[0] }));
  });

  it('burnFrom: should successfully burn tokens, removing them from total supply', async() => {
    await T.approve(accounts[1], 1000, { from: accounts[0] })
    await T.burnFrom(accounts[0], 10, { from: accounts[1] });
    const balance = await T.balanceOf.call(accounts[0]);
    const totalSupply = await T.totalSupply.call();
    const allowance = await T.allowance.call(accounts[0], accounts[1])

    assertNumberIsDifference(balance, balance0, 10);
    assertNumberIsDifference(totalSupply, supply, 10);
    assertNumberIsDifference(allowance, 1000, 10);
  });

  it('burnFrom: revert when paused', async() => {
    await T.approve(accounts[1], 1000, { from: accounts[0] })
    await T.pause({from: owner});
    await assertRevert(T.burnFrom(accounts[0], 10, { from: accounts[1] }));
  });

  it('burnFrom: should successfully burn tokens after unpausing', async() => {
    await T.approve(accounts[1], 1000, { from: accounts[0] })
    await T.pause({from: owner});
    await T.unpause({from: owner});
    await T.burnFrom(accounts[0], 10, { from: accounts[1] });
    const balance = await T.balanceOf.call(accounts[0]);
    const totalSupply = await T.totalSupply.call();
    const allowance = await T.allowance.call(accounts[0], accounts[1])

    assertNumberIsDifference(balance, balance0, 10);
    assertNumberIsDifference(totalSupply, supply, 10);
    assertNumberIsDifference(allowance, 1000, 10);
  });

  it('burnFrom: should successfully burn all tokens in an account', async() => {
    await T.approve(accounts[1], balance0, { from: accounts[0] })
    await T.burnFrom(accounts[0], balance0, { from: accounts[1] });
    const balance = await T.balanceOf.call(accounts[0]);
    const totalSupply = await T.totalSupply.call();
    const allowance = await T.allowance.call(accounts[0], accounts[1])

    assertNumbersEqual(balance, 0);
    assertNumberIsDifference(totalSupply, supply, balance0);
    assertNumbersEqual(allowance, 0);
  });

  it('burnFrom: should successfully burn 0 tokens', async() => {
    await T.approve(accounts[1], 1000, { from: accounts[0] })
    await T.burnFrom(accounts[0], 0, { from: accounts[1] });
    const balance = await T.balanceOf.call(accounts[0]);
    const totalSupply = await T.totalSupply.call();
    const allowance = await T.allowance.call(accounts[0], accounts[1])

    assertNumbersEqual(balance, balance0);
    assertNumbersEqual(totalSupply, supply);
    assertNumbersEqual(allowance, 1000);
  });

  it('burnFrom: should fail burning more tokens than in an account', async() => {
    await T.approve(accounts[1], balance0 + 1000, { from: accounts[0] })
    await assertRevert(T.burnFrom.call(accounts[0], balance0 + 1, { from: accounts[1] }));
  });

  it('burnFrom: should fail burning more tokens than allowance', async() => {
    await T.approve(accounts[1], 1000, { from: accounts[0] })
    await assertRevert(T.burnFrom.call(accounts[0], 1001, { from: accounts[1] }));
  });

  it('transferOwnership: should reassign owner', async() => {
    const obj = await BaseERC20Token.new(supply, 18, "Token", "T");
    assert.strictEqual(await obj.owner.call(), accounts[0]);
    await obj.transferOwnership(accounts[1]);
    assert.strictEqual(await obj.owner.call(), accounts[1]);
    await obj.transferOwnership(accounts[2], {from: accounts[1]});
    assert.strictEqual(await obj.owner.call(), accounts[2]);
  });

  it('transferOwnership: should revert if called by non-owner', async() => {
    const obj = await BaseERC20Token.new(supply, 18, "Token", "T");
    await assertRevert(obj.transferOwnership.call(accounts[1], {from: accounts[2]}));
    await obj.transferOwnership(accounts[1]);
    await assertRevert(obj.transferOwnership.call(accounts[2], {from: accounts[0]}));
  });

  it('renounceOwnership: should remove owner', async() => {
    var obj = await BaseERC20Token.new(supply, 18, "Token", "T");
    assert.strictEqual(await obj.owner.call(), accounts[0]);
    await obj.renounceOwnership();
    assert.strictEqual(await obj.owner.call(), padAddress(0));

    obj = await BaseERC20Token.new(supply, 18, "Token", "T", {from: accounts[1]});
    assert.strictEqual(await obj.owner.call(), accounts[1]);
    await obj.renounceOwnership({from: accounts[1]});
    assert.strictEqual(await obj.owner.call(), padAddress(0));
  });

  it('renounceOwnership: should revert if called by non-owner', async() => {
    var obj = await BaseERC20Token.new(supply, 18, "Token", "T");
    await assertRevert(obj.renounceOwnership.call({from: accounts[1]}));

    obj = await BaseERC20Token.new(supply, 18, "Token", "T", {from: accounts[1]});
    await assertRevert(obj.renounceOwnership.call({from: accounts[0]}));
  });

  it('events: should fire Burn event normally', async () => {
    const res = await T.burn(10, { from: accounts[0] });
    const transferLog = res.logs.find(element => element.event.match('Burn'));
    assert.strictEqual(transferLog.args.burner, accounts[0]);
    assert.strictEqual(transferLog.args.value.toString(), '10');
  });

  it('events: should fire Burn event normally for a 0-burn', async () => {
    const res = await T.burn(0, { from: accounts[0] });
    const transferLog = res.logs.find(element => element.event.match('Burn'));
    assert.strictEqual(transferLog.args.burner, accounts[0]);
    assert.strictEqual(transferLog.args.value.toString(), '0');
  });

  it('events: should fire Pause when paused', async () => {
    const res = await T.pause({ from: owner });
    const transferLog = res.logs.find(element => element.event.match('Pause'));
    assert.ok(!!transferLog, 'should find the Pause event')
  });

  it('events: should fire Unpause when paused', async () => {
    await T.pause({ from: owner });
    const res = await T.unpause({ from: owner });
    const transferLog = res.logs.find(element => element.event.match('Unpause'));
    assert.ok(!!transferLog, 'should find the Unpause event')
  });
  */

  const testTransferEther = async (amount) => {
    const userBalance = await web3.eth.getBalance(accounts[2]);
    const depositBalance = await web3.eth.getBalance(depositAccount.address);

    await depositAccount.transferEther(accounts[2], amount, nonce, {from: depositOwner});

    const userBalanceAfter = await web3.eth.getBalance(accounts[2]);
    const depositBalanceAfter = await web3.eth.getBalance(depositAccount.address);

    assertNumberIsSum(userBalanceAfter, userBalance, amount);
    assertNumberIsDifference(depositBalanceAfter, depositBalance, amount);
  }

  const testTransferEtherToOwner = async (amount) => {
    const userBalance = await web3.eth.getBalance(depositOwner);
    const depositBalance = await web3.eth.getBalance(depositAccount.address);

    await depositAccount.transferEtherToOwner(amount, nonce, {from: depositOwner});

    const userBalanceAfter = await web3.eth.getBalance(depositOwner);
    const depositBalanceAfter = await web3.eth.getBalance(depositAccount.address);

    assertNumberIsGte(userBalanceAfter, BigNumber(userBalance).plus(amount).minus('1e16'));
    assertNumberIsDifference(depositBalanceAfter, depositBalance, amount);
  }

  const testTransferTokens = async (amount) => {
    const userBalance = await token.balanceOf.call(accounts[2]);
    const depositBalance = await token.balanceOf.call(depositAccount.address);

    await depositAccount.transferTokens(token.address, accounts[2], amount, nonce, {from: depositOwner});

    const userBalanceAfter = await token.balanceOf.call(accounts[2]);
    const depositBalanceAfter = await token.balanceOf.call(depositAccount.address);

    assertNumberIsSum(userBalanceAfter, userBalance, amount);
    assertNumberIsDifference(depositBalanceAfter, depositBalance, amount);
  }

  const testTransferTokensToOwner = async (amount) => {
    const userBalance = await token.balanceOf.call(depositOwner);
    const depositBalance = await token.balanceOf.call(depositAccount.address);

    await depositAccount.transferTokensToOwner(token.address, amount, nonce, {from: depositOwner});

    const userBalanceAfter = await token.balanceOf.call(depositOwner);
    const depositBalanceAfter = await token.balanceOf.call(depositAccount.address);

    assertNumberIsSum(userBalanceAfter, userBalance, amount);
    assertNumberIsDifference(depositBalanceAfter, depositBalance, amount);
  }

  const testTransferEtherRecords = async (to, amount, nonce) => {
    const nonceUsed = await depositAccount.nonceUsed.call(nonce);
    const transfer = await depositAccount.transfers.call(nonce);

    assert.strictEqual(nonceUsed, true);
    assert.strictEqual(transfer.token, padAddress(0));
    assert.strictEqual(transfer.to, to);
    assert.strictEqual(transfer.value.toString(), BigNumber(amount).toString());
    assert.strictEqual(transfer.nonce.toString(), BigNumber(nonce).toString());
  }

  const testTransferTokenRecords = async (token, to, amount, nonce) => {
    const nonceUsed = await depositAccount.nonceUsed.call(nonce);
    const transfer = await depositAccount.transfers.call(nonce);

    assert.strictEqual(nonceUsed, true);
    assert.strictEqual(transfer.token, token.address);
    assert.strictEqual(transfer.to, to);
    assert.strictEqual(transfer.value.toString(), BigNumber(amount).toString());
    assert.strictEqual(transfer.nonce.toString(), BigNumber(nonce).toString());
  }

  const testTransferEtherLogs = async (res, to, amount, nonce) => {
    const transferLog = res.logs.find(element => element.event === 'Transferred');
    assert.ok(!!transferLog, 'should find the Transferred event')

    assert.strictEqual(transferLog.args.token, padAddress(0));
    assert.strictEqual(transferLog.args.to, to);
    assertNumbersEqual(transferLog.args.value, amount);
    assert.strictEqual(transferLog.args.nonce.toString(), BigNumber(nonce).toString());

    const transferEtherLog = res.logs.find(element => element.event === 'TransferredEther');
    assert.ok(!!transferEtherLog, 'should find the TransferredEther event');

    assert.strictEqual(transferEtherLog.args.to, to);
    assertNumbersEqual(transferEtherLog.args.value, amount);
    assert.strictEqual(transferEtherLog.args.nonce.toString(), BigNumber(nonce).toString());

    const transferTokenLog = res.logs.find(element => element.event === 'TransferredTokens');
    assert.ok(!transferTokenLog, 'should not find the TransferredTokens event');
  }

  const testTransferTokenLogs = async (res, token, to, amount, nonce) => {
    const transferLog = res.logs.find(element => element.event === 'Transferred');
    assert.ok(!!transferLog, 'should find the Transferred event')

    assert.strictEqual(transferLog.args.token, token.address);
    assert.strictEqual(transferLog.args.to, to);
    assertNumbersEqual(transferLog.args.value, amount);
    assert.strictEqual(transferLog.args.nonce.toString(), BigNumber(nonce).toString());

    const transferTokenLogs = res.logs.find(element => element.event === 'TransferredTokens');
    assert.ok(!!transferTokenLogs, 'should find the TransferredTokens event');

    assert.strictEqual(transferTokenLogs.args.token, token.address);
    assert.strictEqual(transferTokenLogs.args.to, to);
    assertNumbersEqual(transferTokenLogs.args.value, amount);
    assert.strictEqual(transferTokenLogs.args.nonce.toString(), BigNumber(nonce).toString());

    const transferEtherLog = res.logs.find(element => element.event === 'TransferredEther');
    assert.ok(!transferEtherLog, 'should not find the TransferredEther event');
  }

  const sendEther = async (from, to, value) => {
    return new Promise((resolve, reject) => web3.eth.sendTransaction({from, to, value, gas: 100000}, (e,res) => (e ? reject(e) : resolve(res))));
  }
})
